/* Copyright 2000-2001 Red Hat, Inc.  All Rights Reserved.
 * Written by Benjamin LaHaise <bcrl@redhat.com>
 * This file may be distributed under the terms of the
 * GNU Public License v2 or later, but is provided with NO WARRANTY.
 *
 * asyncbuf2: aio test program that provides an elastic buffer between
 * stdin and stdout.
 */
#include <aio.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>

#define BUFSIZE		(4 * 1024 * 1024)

struct stream {
	char		buffer[BUFSIZE];
	
	loff_t		rpos;
	loff_t		wpos;

	int		in_fd;
	int		out_fd;

	int		in_done, out_done;

	struct iocb	read_iocb;
	struct iocb	write_iocb;
};

/* borrowed from linux/list.h */
#define struct_from_entry(ptr, type, member) \
        ((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))


void read_callback(aio_context_t ctx, struct iocb *iocb, long res, long res2)
{
	struct stream *s = struct_from_entry(iocb, struct stream, read_iocb);
	if (res < 0) {
		fprintf(stderr, "write: %ld %ld\n", res, res2);
		exit(1);
	} else if (res) {
		s->rpos += res;
	} else
		s->in_done = 1;
}

void write_callback(aio_context_t ctx, struct iocb *iocb, long res, long res2)
{
	if (res <= 0) {
		fprintf(stderr, "write: %ld %ld\n", res, res2);
		exit(1);
	}
}


int main (int argc, char *argv[])
{
	struct aiocb read_iocb, write_iocb;
	struct aiocb *aio_list[2];
	unsigned long rpos = 0, wpos = 0, buf_bytes = 0;
	off_t roff = 0, woff = 0;
	int read_active = 0, write_active = 0;
	int saw_eof = 0;
	int in_fd = 0, out_fd = 1;
	int i;
	ssize_t ret;
	time_t then = time(NULL);
	int read_wait = 0;

	buffer = foobuffer + PAGE_SIZE - ((unsigned long)foobuffer & (PAGE_SIZE - 1));

	if (argc == 3) {
		fprintf(stderr, "using O_SYNC\n");

		in_fd = open(argv[1], O_RDONLY);
		if (in_fd < 0) {
			perror("open() input");
			return 1;
		}
		out_fd = open(argv[2], O_WRONLY | O_SYNC | O_CREAT, 0644);
		if (out_fd < 0) {
			perror("open() output");
			return 1;
		}
	} else if (argv != 1) {
		fprintf(stderr, "usage: asyncbuf infile outfile  or asyncbuf\n");
		return 2;
	}

	memset(&read_iocb, 0, sizeof(read_iocb));
	memset(&write_iocb, 0, sizeof(write_iocb));

	do {
		if (!write_active && buf_bytes) {
			write_iocb.aio_sigevent.sigev_notify = SIGEV_NONE;
			write_iocb.aio_offset = woff;
			write_iocb.aio_fildes = out_fd;
			write_iocb.aio_buf = buffer + wpos;
			write_iocb.aio_nbytes = buf_bytes;
			if ((buf_bytes >> 2) > WRITESIZE)
				write_iocb.aio_nbytes >>= 2;

			if (wpos + write_iocb.aio_nbytes >= BUFSIZE)
				write_iocb.aio_nbytes = BUFSIZE - wpos;

			do
				ret = aio_write(&write_iocb);
			while (ret && errno == EINTR);

			if (ret)
				goto write_err;

			write_active = 1;
		}

		if (read_wait && (BUFSIZE - buf_bytes) >= MINREADSPACE)
			read_wait = 0;

		if (!read_wait && !read_active && buf_bytes < BUFSIZE && !saw_eof) {
			read_iocb.aio_sigevent.sigev_notify = SIGEV_NONE;
			read_iocb.aio_offset = roff;
			read_iocb.aio_fildes = in_fd;
			read_iocb.aio_buf = buffer + rpos;
			read_iocb.aio_nbytes = BUFSIZE - buf_bytes;

			if (read_iocb.aio_nbytes > READSIZE)
				read_iocb.aio_nbytes = READSIZE;
				
			if (rpos + read_iocb.aio_nbytes >= BUFSIZE)
				read_iocb.aio_nbytes = BUFSIZE - rpos;

			do
				ret = aio_read(&read_iocb);
			while (ret && errno == EINTR);

			if (ret)
				goto read_err;
			read_active = 1;
		}

		i = 0;
		if (write_active)
			aio_list[i++] = &write_iocb;
		if (read_active)
			aio_list[i++] = &read_iocb;

		ret = aio_suspend(&aio_list, i, NULL);
		if (ret)
			goto err;

		if (read_active && (ret = aio_error(&read_iocb)) != EINPROGRESS) {
			read_active = 0;
			if (ret)
				goto read_err;

			ret = aio_return(&read_iocb);
			if (ret < 0)
				goto read_err;

			buf_bytes += ret;
			roff += ret;
			rpos += ret;
			rpos %= BUFSIZE;
			if (!ret)
				saw_eof = 1;

			if (buf_bytes == BUFSIZE)
				read_wait = 1;
		}

		if (write_active && (ret = aio_error(&write_iocb)) != EINPROGRESS) {
			write_active = 0;
			if (ret)
				goto write_err;

			ret = aio_return(&write_iocb);
			if (ret < 0)
				goto write_err;

			buf_bytes -= ret;
			woff += ret;
			wpos += ret;
			wpos %= BUFSIZE;
			if (!ret)
				break;
		}

		if (time(NULL) - then >= 1) {
			then = time(NULL);
			fprintf(stderr, "\r%3ld%%", (long)(buf_bytes * 100) / BUFSIZE);
		}
#if 0
fprintf(stderr, "wpos: %ld rpos: %ld buf_bytes: %ld saw_eof: %d i: %d\n",
	wpos, rpos, buf_bytes, saw_eof, i);
#endif
	} while (i);

	return 0;
err:
	perror("aio_suspend");
	fprintf(stderr, "ret: %d\n", ret);

	return 1;

read_err:
	perror("aio_read");
	fprintf(stderr, "ret: %d\n", ret);
	return 1;

write_err:
	perror("aio_write");
	fprintf(stderr, "ret: %d\n", ret);
	return 1;

}

